// CPerformanceTimer - a simple Win32 performance counter wrapper
// by Dean Wyant dwyant@mindspring.com

/**
 * \file PerformanceTimer.h
 * A simple Win32 performance counter wrapper, by Dean Wyant dwyant@mindspring.com
 */

/*

  This class is simple to use. Just declare a variable(s) as type CPerformanceTimer,
  call Start() to start timimg and call Stop() to stop timimg. You can pause a
  timer by calling Stop() and then you can call Start() to resume. Retrieve the
  elapsed time by calling an Elapsed..() function. Assignment, addition,
  subtraction and comparison are supported. There are a few information calls
  available also. All calls except Start and Stop can be performed on a timer
  without stopping it.

*/

#ifndef __PERFTIMER_H__
#define __PERFTIMER_H__

#include <windows.h>

/**
 * class interface to QueryPerformanceCounter.
 */
class CPerformanceTimer
{
public:
  CPerformanceTimer( BOOL bStart = FALSE ) { Init(bStart); }

  CPerformanceTimer(const CPerformanceTimer& Src);

  virtual ~CPerformanceTimer() {;}

  void Start(BOOL bReset = FALSE);   // Start from current value or optionally from 0
  void Stop();                       // Stop timing. Use Start afterwards to continue.

  BOOL IsRunning();                  // Returns FALSE if stopped.

  BOOL IsSupported();                // Returns FALSE if performance counter not supported.
                                     // Call after constructing at least one CPerformanceTimer

  const double Resolution();         // Returns timer resolution in seconds
  const double Resolutionms();       // Returns timer resolution in milliseconds
  const double Resolutionus();       // Returns timer resolution in microseconds

  const double Elapsed();            // Returns elapsed time in seconds
  const double Elapsedms();          // Returns elapsed time in milliseconds
  const double Elapsedus();          // Returns elapsed time in microseconds

  const CPerformanceTimer& operator=(const CPerformanceTimer& Src); // Assignment operator

  // Math operators
  CPerformanceTimer operator+(const CPerformanceTimer& Src) const;
	CPerformanceTimer operator-(const CPerformanceTimer& Src) const;
	const CPerformanceTimer& operator+=(const CPerformanceTimer& Src);
	const CPerformanceTimer& operator-=(const CPerformanceTimer& Src);
  // For time in seconds
  CPerformanceTimer operator+(const double Secs) const;
	CPerformanceTimer operator-(const double Secs) const;
	const CPerformanceTimer& operator+=(const double Secs);
	const CPerformanceTimer& operator-=(const double Secs);

  // Boolean comparison operators
	BOOL operator<(const CPerformanceTimer& Src);
	BOOL operator>(const CPerformanceTimer& Src);
	BOOL operator<=(const CPerformanceTimer& Src);
	BOOL operator>=(const CPerformanceTimer& Src);
  // For time in seconds
  BOOL operator<(const double Secs);
	BOOL operator>(const double Secs);
	BOOL operator<=(const double Secs);
	BOOL operator>=(const double Secs);

  virtual void Lock() const {;}     // Override for thread safe operation
  virtual void Unlock() const {;}     // Override for thread safe operation
protected:
  void Init(BOOL bStart);
  void Copy(const CPerformanceTimer& Src);

private:
  __int64 m_Start;
  static __int64 m_Freq;   // does not change while system is running
  static __int64 m_Adjust; // Adjustment time it takes to Start and Stop
};

/**
 * You only need to use types of this class if a timer is going to be shared between threads.
 */
class CPerfTimerT : public CPerformanceTimer
{ // You only need to use types of this class if a timer is going to be shared between threads
public:
  CPerfTimerT(BOOL bStart = FALSE)
  {
    m_hMutex = CreateMutex(NULL,FALSE,"");
    Init(bStart);
  }

  CPerfTimerT(const CPerfTimerT& Src)
  {
    m_hMutex = CreateMutex(NULL,FALSE,"");
    Copy(Src);
  }

  CPerfTimerT(const CPerformanceTimer& Src)
  {
    m_hMutex = CreateMutex(NULL,FALSE,"");
    Copy(Src);
  }

  virtual ~CPerfTimerT()
  { CloseHandle(m_hMutex); }

  const CPerfTimerT& operator=(const CPerfTimerT& Src) // Assignment operator
  {
    Copy(Src);
    return *this;
  }

  virtual void Lock() const { WaitForSingleObject(m_hMutex,10000); }
  virtual void Unlock() const { ReleaseMutex(m_hMutex); }
private:
  HANDLE m_hMutex;
};

inline void CPerformanceTimer::Init(BOOL bStart)
{
  if (!m_Freq)
  { // Initialization should only run once
    QueryPerformanceFrequency((LARGE_INTEGER *)&m_Freq);
    if (!m_Freq)
      m_Freq = 1; // Timer will be useless but will not cause divide by zero
    m_Start = 0;
    m_Adjust = 0;
    Start();            // Time a Stop
    Stop();
    m_Adjust = m_Start;
  }
  // This is the only part that normally runs
  m_Start = 0;
  if (bStart)
    Start();
}

inline CPerformanceTimer::CPerformanceTimer(const CPerformanceTimer& Src)
{
  Copy(Src);
}

inline void CPerformanceTimer::Copy(const CPerformanceTimer& Src)
{
  if (&Src == this)
    return; // avoid deadlock if someone tries to copy it to itself
  Src.Lock();
  Lock();
  m_Start = Src.m_Start;
  Unlock();
  Src.Unlock();
}

inline void CPerformanceTimer::Start(BOOL bReset)
{ // Start from current value or optionally from 0
  __int64 i;
  QueryPerformanceCounter((LARGE_INTEGER *)&i);
  Lock();
  if ((!bReset) && (m_Start < 0))
    m_Start += i;   // We are starting with an accumulated time
  else
    m_Start = i;    // Starting from 0
  Unlock();
}

inline void CPerformanceTimer::Stop()
{ // Stop timing. Use Start afterwards to continue
  Lock();
  if (m_Start <= 0)
  {
    Unlock();
    return;          // Was not running
  }
  __int64 i;
  QueryPerformanceCounter((LARGE_INTEGER *)&i);
  m_Start += -i;          // Stopped timer keeps elapsed timer ticks as a negative
  if (m_Start < m_Adjust) // Do not overflow
    m_Start -= m_Adjust;  // Adjust for time timer code takes to run
  else
    m_Start = 0;          // Stop must have been called directly after Start
  Unlock();
}

inline BOOL CPerformanceTimer::IsRunning()
{ // Returns FALSE if stopped.
  Lock();
  BOOL bRet = (m_Start > 0); // When < 0, holds elpased clicks
  Unlock();
  return bRet;
}
 inline const double CPerformanceTimer::Elapsed()
{ // Returns elapsed time in seconds
  CPerformanceTimer Result(*this);
  Result.Stop();
  return (double)(-Result.m_Start)/(double)m_Freq;
}

inline const double CPerformanceTimer::Elapsedms()
{ // Returns elapsed time in milliseconds
  CPerformanceTimer Result(*this);
  Result.Stop();
  return (-Result.m_Start*1000.0)/(double)m_Freq;
}

inline const double CPerformanceTimer::Elapsedus()
{ // Returns elapsed time in microseconds
  CPerformanceTimer Result(*this);
  Result.Stop();
  return (-Result.m_Start * 1000000.0)/(double)m_Freq;
}


// Assignment operator
inline const CPerformanceTimer& CPerformanceTimer::operator=(const CPerformanceTimer& Src)
{
  Copy(Src);
  return *this;
}


// Math operators
inline CPerformanceTimer CPerformanceTimer::operator+(const CPerformanceTimer& Src) const
{
  CPerformanceTimer Result(*this);
  Result += Src;
  return Result;
}

inline CPerformanceTimer CPerformanceTimer::operator-(const CPerformanceTimer& Src) const
{
  CPerformanceTimer Result(*this);
  Result -= Src;
  return Result;
}

inline const CPerformanceTimer& CPerformanceTimer::operator+=(const CPerformanceTimer& Src)
{
  CPerformanceTimer SrcStop(Src);  // Temp is necessary in case Src is not stopped
  SrcStop.Stop();
  Lock();
  m_Start += SrcStop.m_Start;
  Unlock();
  return *this;
}

inline const CPerformanceTimer& CPerformanceTimer::operator-=(const CPerformanceTimer& Src)
{
  CPerformanceTimer SrcStop(Src);  // Temp is necessary in case Src is not stopped
  SrcStop.Stop();
  Lock();
  m_Start -= SrcStop.m_Start;
  Unlock();
  return *this;
}

// For time in seconds
inline CPerformanceTimer CPerformanceTimer::operator+(const double Secs) const
{
  CPerformanceTimer Result(*this);
  Result += Secs;
  return Result;
}

inline CPerformanceTimer CPerformanceTimer::operator-(const double Secs) const
{
  CPerformanceTimer Result(*this);
  Result += Secs;
  return Result;
}

inline const CPerformanceTimer& CPerformanceTimer::operator+=(const double Secs)
{
  Lock();
  m_Start -= (__int64)(Secs*(double)m_Freq);
  Unlock();
  return *this;
}

inline const CPerformanceTimer& CPerformanceTimer::operator-=(const double Secs)
{
  Lock();
  m_Start += (__int64)(Secs*(double)m_Freq);
  Unlock();
  return *this;
}



// Boolean comparison operators
inline BOOL CPerformanceTimer::operator<(const CPerformanceTimer& Src)
{
  BOOL bRet;
  CPerformanceTimer Temp(Src);
  Lock();
  if (m_Start <= 0)
  {
    Temp.Stop();
    bRet = (m_Start > Temp.m_Start);
    Unlock();
    return bRet;
  }
  else
  if (Temp.m_Start > 0)
  {
    bRet = (m_Start < Temp.m_Start);
    Unlock();
    return bRet;
  }
  else
  {
    Unlock();
    CPerformanceTimer ThisStop(*this);
    ThisStop.Stop();
    return (ThisStop.m_Start > Temp.m_Start);
  }
}

inline BOOL CPerformanceTimer::operator>(const CPerformanceTimer& Src)
{
  BOOL bRet;
  CPerformanceTimer Temp(Src);
  Lock();
  if (m_Start <= 0)
  {
    Temp.Stop();
    bRet = (m_Start < Temp.m_Start);
    Unlock();
    return bRet;
  }
  else
  if (Temp.m_Start > 0)
  {
    bRet = (m_Start > Temp.m_Start);
    Unlock();
    return bRet;
  }
  else
  {
    Unlock();
    CPerformanceTimer ThisStop(*this);
    ThisStop.Stop();
    return (ThisStop.m_Start < Temp.m_Start);
  }
}

inline BOOL CPerformanceTimer::operator<=(const CPerformanceTimer& Src)
{
  return !(*this > Src);
}

inline BOOL CPerformanceTimer::operator>=(const CPerformanceTimer& Src)
{
  return !(*this < Src);
}

// For time in seconds
inline BOOL CPerformanceTimer::operator<(const double Secs)
{
  BOOL bRet;
  Lock();
  if (m_Start <= 0)
  {
    bRet = (m_Start > (__int64)(-Secs*(double)m_Freq));
    Unlock();
    return bRet;
  }
  else
  {
    Unlock();
    CPerformanceTimer ThisStop(*this);
    ThisStop.Stop();
    return (ThisStop.m_Start > (__int64)(-Secs*(double)m_Freq));
  }
}

inline BOOL CPerformanceTimer::operator>(const double Secs)
{
  BOOL bRet;
  Lock();
  if (m_Start <= 0)
  {
    bRet = (m_Start < (__int64)(-Secs*(double)m_Freq));
    Unlock();
    return bRet;
  }
  else
  {
    Unlock();
    CPerformanceTimer ThisStop(*this);
    ThisStop.Stop();
    return (ThisStop.m_Start < (__int64)(-Secs*(double)m_Freq));
  }
}

inline BOOL CPerformanceTimer::operator<=(const double Secs)
{
  return !(*this > Secs);
}

inline BOOL CPerformanceTimer::operator>=(const double Secs)
{
  return !(*this < Secs);
}


/*
 * originally in .cpp file:
 */

inline BOOL CPerformanceTimer::IsSupported()
{ // Returns FALSE if performance counter not supported.
  // Call after constructing at least one CPerformanceTimer
  return (m_Freq > 1);
}

inline const double CPerformanceTimer::Resolution()
{ // Returns timer resolution in seconds
  return 1.0/(double)m_Freq;
}

inline const double CPerformanceTimer::Resolutionms()
{ // Returns timer resolution in milliseconds
  return 1000.0/(double)m_Freq;
}

inline const double CPerformanceTimer::Resolutionus()
{ // Returns timer resolution in microseconds
  return 1000000.0/(double)m_Freq;
}


#endif //__PERFTIMER_H__
