/*
 * Register.h
 *
 * ACCU Overload 95, February 2010
 *
 * One approach to using hardware registers in C++.
 *
 * Martin Moene, http://www.eld.physics.LeidenUniv.nl/~moene/
 */

#ifndef G_REGISTER_H
#define G_REGISTER_H

/**
 * \file Code/Register.h
 * A register class, transport via channel policy; transport for and creation
 * of memory mapped registers provided.
 */

// system includes
#include <memory>                         // std::auto_ptr
#include <boost/noncopyable.hpp>          // boost::noncopyable
#include <boost/static_assert.hpp>        // BOOST_STATIC_ASSERT()
#include <boost/type_traits/is_same.hpp>  // boost::is_same

// project includes
#include "MemoryChannel.h"                // spm::MemoryChannel

/**
 * scanning probe microscopy.
 */
namespace spm {

/**
 * A Register; transport policy (for example memory mapped, or bus) is provided
 * as a template type parameter.
 *
 * Example:
 * \code
 * typedef int DataType;
 * typedef spm::Register< DataType > TRegister;
 * typedef TRegister::AddressType AddressType;
 *
 * int x;  // simulate hardware register
 * const AddressType base  ( &x );
 * const int         offset(  0 );
 *
 * TRegister reg( base, offset );
 *
 * reg = 456;                    // write given value to memory
 * reg.write();                  // write cached value to memory again
 * reg.write_nc( 987 );          // write given value to memory, but do not cache its value
 *
 * bool x = reg.bittest( 3 );    // test if bit 3 is set
 * reg.bitclear( 0 );            // clear first bit
 *
 * reg.maskclear( 0xf00 );       // clear bits according to given mask
 * reg.maskset( 0x0ff, 0x00f );  // clear bits in first mask, set bits in second mask
 * \endcode
 *
 * How exactly the register is read and written is left to a channel policy CP (or
 * transport policy).
 * \code
 * template
 * < typename D
 * , typename A = volatile D*
 * , typename CP = spm::MemoryChannel<D,A>
 * >
 * class Register : public boost::noncopyable;
 * \endcode
 */
template
< typename D
, typename A = volatile D*
, typename CP = MemoryChannel<D,A>
>
class Register : public boost::noncopyable
{
public:
   /**
    * the register data type.
    */
   typedef D DataType;

   /**
    * the register address type.
    */
   typedef A AddressType;

   /**
    * the transport channel type.
    */
   typedef CP ChannelType;

   /**
    * this type.
    */
   typedef Register Self;

   /**
    * constructor without channel object parameter.
    *
    * Note that although a register just has an address and the offset only makes
    * sense in relation the a card's base address, offset is included here to
    * confine the register's address computation to a single location.
    */
   Register( AddressType address, int offset, DataType data = 0 )
   : m_channel_smartptr()           // Note: declared before m_channel
   , m_channel( createChannel() )
   , m_address( computeAddress( address, offset ) )
   , m_cache( data )
   {
      ; // do nothing
   }

   /**
    * constructor with channel object parameter.
    *
    * Note that although a register just has an address and the offset only makes
    * sense in relation the a card's base address, offset is included here to
    * confine the register's address computation to a single location.
    */
   Register( ChannelType& channel, AddressType address, int offset, DataType data = 0 )
   : m_channel_smartptr()           // Note: declared before m_channel
   , m_channel( channel )
   , m_address( computeAddress( address, offset ) )
   , m_cache  ( data    )
   {
      ; // do nothing
   }

   /**
    * destructor: used for type checking.
    */
   ~Register()
   {
      checkTypes();
   }

   /**
    * ensure that data and address types of channel policy and register match.
    */
   static void checkTypes()
   {
      BOOST_STATIC_ASSERT( ( boost::is_same<D, typename CP::DataType>::value ) );
      BOOST_STATIC_ASSERT( ( boost::is_same<A, typename CP::AddressType>::value ) );
   }

   /**
    * read data and return it; same as read().
    */
   operator DataType()
   {
      return read();
   }

   /**
    * write data and return self; in effect write(data).
    */
   Self& operator= ( DataType data )
   {
      write( data );
      return *this;
   }

   /**
    * reinitialize the register on the card from the cached value.
    */
   void reinitialize()
   {
      write();
   }

   /**
    * the register's address.
    */
   AddressType address() const
   {
      return m_address;
   }

   /**
    * the register's cached data (non-reading).
    */
   DataType cache() const
   {
      return m_cache;
   }

   /**
    * read the register and return its value (reading).
    */
   DataType read()
   {
      return m_channel.read( m_address );
   }

   /**
    * write the cached value to the register on the card.
    */
   void write()
   {
      m_channel.write( m_address, m_cache );
   }

   /**
    * write the given value to the register (caching).
    */
   void write( DataType value )
   {
      m_channel.write( m_address, m_cache = value );
   }

   /**
    * write the given value to the register (non-caching).
    */
   void write_nc( DataType value )
   {
      m_channel.write( m_address, value );
   }

   /**
    * test the bit at the given position of the register (reading).
    */
   bool bittest( int bit )
   {
      return 0 != ( read() & bitmask( bit ) );
   }

   /**
    * clear the bit at the given position of the register (cached).
    */
   void bitclear( int bit )
   {
      write( m_cache & ~bitmask( bit ) );
   }

   /**
    * set the bit at the given position of the register (cached).
    */
   void bitset( int bit )
   {
      write( m_cache | bitmask( bit ) );
   }

   /**
    * test the bits in the register at the positions given by the mask (reading).
    */
   bool masktest( DataType mask )
   {
      return mask == ( read() & mask );
   }

   /**
    * clear the bits in the register at the positions given by the mask (caching).
    */
   void maskclear( DataType mask )
   {
      write( m_cache & ~mask );
   }

   /**
    * set the bits in the register at the positions given by the mask (caching).
    */
   void maskset( DataType mask )
   {
      write( m_cache | mask );
   }

   /**
    * in the register, clear the bits given by clearmask, set the bits given by the setmask (caching).
    */
   void maskset( DataType clearmask, DataType setmask )
   {
      m_cache &= ~clearmask;
      maskset( setmask );
   }

   /**
    * create bit mask for given bit.
    */
   static DataType bitmask( int bit )
   {
      return 1 << bit;
   }

   /**
    * compute register address method.
    */
   static AddressType computeAddress( AddressType base, int offset )
   {
      return base + offset;
   }

private:
   /**
    * constructor helper method.
    */
   ChannelType& createChannel()
   {
      m_channel_smartptr.reset( new ChannelType() );
      return *m_channel_smartptr;
   }

private:
   /*
    * Note1: std::auto_ptr must be declared before m_channel (init before use).
    * Note2: std::auto_ptr is to be replaced by std::unique_ptr (C++0X).
    */

   /**
    * smart pointer for internally created channels, such as transport for memory mapped registers.
    */
   std::auto_ptr<ChannelType> m_channel_smartptr;

   /**
    * the transport channel policy.
    */
   ChannelType& m_channel;

   /**
    * the register's address.
    */
   AddressType m_address;

   /**
    * the cached written data.
    */
   DataType m_cache;
};

} // namespace spm {

#endif // G_REGISTER_H

/*
 * end of file
 */
