/*
 * RegisterSpy.h
 *
 * ACCU Overload 95, February 2010
 *
 * One approach to using hardware registers in C++.
 *
 * Martin Moene, http://www.eld.physics.LeidenUniv.nl/~moene/
 */

#ifndef G_REGISTERSPY_H
#define G_REGISTERSPY_H

/**
 * \file RegisterSpy.h
 * A configurable register spy.
 */

// system includes
#include <algorithm>    // std::for_each
#include <functional>   // std::unary_function
#include <iomanip>      // std::setw() etc.
#include <iosfwd>       // std::ostream
#include <string>       // std::string
#include <vector>       // std::vector<>

// project includes
#include "Cstdint.h"    // std::uint32_t
#include "Shims.h"      // spm::to_string()

/**
 * match expectations and occurrences.
 */
#define  SPM_CHECK_REGISTER_SPY_EXPECTATIONS( spy ) \
   SPM_CHECK_EQUAL_COLLECTIONS( spy.getAllExpectations(), spy.getAllOccurrences() ) \

/**
 * match expectations and occurrences.
 */
#define  SPM_CHECK_EQUAL_COLLECTIONS( exp, occ ) \
   BOOST_CHECK_EQUAL_COLLECTIONS( exp.begin(), exp.end(), occ.begin(), occ.end() )

/**
 * scanning probe microscopy.
 */
namespace spm {

/**
 * inspection driven design.
 */
namespace tdd {

/**
 * configurable register spy (implements Channel Concept).
 */
template < typename D, typename A >
class RegisterSpy
{
public:
   /**
    * data type.
    */
   typedef D DataType;

   /**
    * address type.
    */
   typedef A AddressType;

   /**
    * this type.
    */
   typedef RegisterSpy<D,A> Self;

   /**
    * the operation type.
    */
   enum Operation
   {
      eOp_Read,   ///< a read operation
      eOp_Write,  ///< a write operation
   };

   /**
    * the expectation type.
    */
   struct Expectation
   {
      typedef Expectation Self;  ///< this type

      Operation    operation;    ///< the operation
      AddressType  address;      ///< the address
      DataType     data;         ///< the data
      std::string  remark;       ///< action description

      /**
       * default constructor.
       */
      Expectation()
      : operation(), address(), data(), remark()
      {
         ; // do nothing
      }

      /**
       * constructor.
       */
      Expectation( Operation operation_, AddressType address_, DataType data_, std::string remark_ = "" )
      : operation( operation_ ), address( address_ ), data( data_ ), remark( remark_ )
      {
         ; // do nothing
      }

      /**
       * equality.
       */
      bool operator== ( Self const& rhs ) const
      {
         return operation == rhs.operation
             &&   address == rhs.address
             &&      data == rhs.data;
      }

      /**
       * inequality; required by Boost.Test.
       */
      bool operator!= ( Self const& rhs ) const
      {
         return ! ( *this == rhs );
      }

      /**
       * stream expectation output operator (compact representation).
       */
      friend std::ostream& operator<< ( std::ostream& os, Expectation const& exp )
      {
         return exp.streamCompact( os );
      }

      /**
       * stream expectation (compact representation).
       */
      std::ostream& streamCompact( std::ostream& os ) const
      {
         std::uint32_t uladdress( (std::uint32_t) address );

         return os << "[" << (eOp_Read == operation ? "read,":"write,")
                   << "0x" << std::hex << uladdress << ","
                   << "0x" << std::hex << data      << "]";
      }

      /**
       * stream expectation (full representation).
       */
      std::ostream& streamFull( std::ostream& os ) const
      {
         const std::uint32_t uladdress( (std::uint32_t) address );
         const std::string   tabbar   ( remark.length() ? "\t| " : "" );

         return os <<   "0x" << std::hex << std::setfill('0') << std::setw(4) << uladdress
                   << (operation==eOp_Write ? " <-- ":" --> ")
                   <<   "0x" << std::hex << std::setfill('0') << std::setw(4) << data
                   << "  0b" << std::setw(16) << spm::to_string( data, 2).c_str()
                   <<   "  " << std::dec << data
                   << tabbar << remark   << std::endl;
      }
   };

   /**
    * stream expectation function object (sequence count, full representation).
    */
   struct streamExpectation : public std::unary_function< Expectation, Expectation >
   {
      std::ostream& os;    ///< the output stream
      std::string prefix;  ///< the prefix
      int count;           ///< sequence number

      /**
       * constructor.
       */
      streamExpectation( std::ostream& os_, std::string prefix_ = "" )
      : os    ( os_     )
      , prefix( prefix_ )
      , count ( 0       )
      {
         ; // do nothing
      }

      /**
       * apply.
       */
      void operator()( Expectation const& exp )
      {
         os << prefix << std::dec << std::setfill(' ') << std::setw(3) << ++count << ": ";
         (void) exp.streamFull( os );
      }
   };

   /**
    * vector of expectations type.
    */
   typedef std::vector< Expectation > TVecExpectation;

   /**
    * (default) constructor.
    */
   RegisterSpy( std::string prefix = "" )
   : m_prefix( prefix )
   , m_relax( false )
   , m_current( 0 )
   {
      ; // do nothing
   }

   /**
    * stream all occurrences to given stream (full representation).
    */
   friend std::ostream& operator<< ( std::ostream& os, Self const& exp )
   {
      return exp.streamOccurrences( os );
   }

   /**
    * stream all expectatons to given stream (full representation).
    */
   std::ostream& streamExpectations( std::ostream& os ) const
   {
      return streamCollection( os, getAllExpectations() );
   }

   /**
    * stream all occurrences to given stream (full representation).
    */
   std::ostream& streamOccurrences( std::ostream& os ) const
   {
      return streamCollection( os, getAllOccurrences() );
   }

   /**
    * clear the collected information.
    */
   void clear()
   {
      m_relax = false;
      m_current = 0;
      m_vecExpectation.clear();
      m_vecOccurrence.clear();
   }

   /**
    * remove requirement to match each read and write with an expectation;
    * useful to only inspect occurrences.
    */
   void relax( bool relax = true )
   {
      m_relax = relax;
   }

   /**
    * get address at given index.
    */
   AddressType getExpectedOperation( int n ) const
   {
      return m_vecExpectation[ n ].operation;
   }

   /**
    * get address at given index.
    */
   AddressType getExpectedAddress( int n ) const
   {
      return m_vecExpectation[ n ].address;
   }

   /**
    * get data at given index.
    */
   DataType getExpectedData( int n ) const
   {
      return m_vecExpectation[ n ].data;
   }

   /**
    * get address at given index.
    */
   AddressType getActualOperation( int n ) const
   {
      return m_vecOccurrence[ n ].operation;
   }

   /**
    * get address at given index.
    */
   AddressType getActualAddress( int n ) const
   {
      return m_vecOccurrence[ n ].address;
   }

   /**
    * get data at given index.
    */
   DataType getActualData( int n ) const
   {
      return m_vecOccurrence[ n ].data;
   }

   /**
    * get the current expectation tuple.
    */
   Expectation getCurrentExpectation()
   {
      return m_vecExpectation[ m_current++ ];
   }

   /**
    * get given occurrence tuple.
    */
   Expectation getExpectation( int n ) const
   {
      return m_vecExpectation[ n ];
   }

   /**
    * get given occurrence tuple.
    */
   Expectation getOccurrence( int n ) const
   {
      return m_vecOccurrence[ n ];
   }

   /**
    * get all expectation information.
    */
   TVecExpectation const& getAllExpectations() const
   {
      return m_vecExpectation;
   }

   /**
    * get all collected information.
    */
   TVecExpectation const& getAllOccurrences() const
   {
      return m_vecOccurrence;
   }

   /**
    * add a write expectation to the collection.
    */
   void addReadExpectation( AddressType address, DataType data, std::string remark = "" )
   {
      m_vecExpectation.push_back( Expectation( eOp_Read, address, data, remark ) );
   }

   /**
    * add a write expectation to the collection.
    */
   void addWriteExpectation( AddressType address, DataType data, std::string remark = "" )
   {
      m_vecExpectation.push_back( Expectation( eOp_Write, address, data, remark ) );
   }

   /**
    * add a read occurrence to the collection.
    */
   void addReadOccurrence( AddressType address, DataType data, std::string remark = "" )
   {
      m_vecOccurrence.push_back( Expectation( eOp_Read, address, data, remark ) );
   }

   /**
    * add a write occurrence to the collection.
    */
   void addWriteOccurrence( AddressType address, DataType data, std::string remark = "" )
   {
      m_vecOccurrence.push_back( Expectation( eOp_Write, address, data, remark ) );
   }

   /**
    * \name Channel concept.
    * @{
    */

   /**
    * read.
    */
   DataType read( AddressType address )
   {
      Expectation exp = m_relax ? Expectation() : getCurrentExpectation();

      addReadOccurrence( address, exp.data, exp.remark );
      return exp.data;
   }

   /**
    * read.
    */
   void write( AddressType address, DataType data )
   {
      std::string remark;

      if ( !m_relax )
      {
         remark = getCurrentExpectation().remark;
      }
      addWriteOccurrence( address, data, remark );
   }
   /// @} channel concept

private:
   /**
    * stream given expectations to given stream.
    */
   std::ostream& streamCollection( std::ostream& os, TVecExpectation const& vecExp ) const
   {
      std::for_each(
         vecExp.begin(),
         vecExp.end(),
         streamExpectation( os, m_prefix )
      );
      return os;
   }

private:
   /**
    * stream output prefix.
    */
   std::string m_prefix;

   /**
    * relax the expectation--occurrence symmetry.
    */
   int m_relax;

   /**
    * the expectation currently exercised.
    */
   int m_current;

   /**
    * the expectations set.
    */
   TVecExpectation m_vecExpectation;

   /**
    * the occurrences obtained.
    */
   TVecExpectation m_vecOccurrence;
};

} } // namespace spm::tdd

#endif // G_REGISTERSPY_H

/*
 * end of file
 */
