/*
 * Test-performance.cpp
 *
 * ACCU Overload 95, February 2010
 *
 * One approach to using hardware registers in C++.

 * Martin Moene, http://www.eld.physics.LeidenUniv.nl/~moene/
 */

/**
 * \file Test-performance.cpp
 * Time performance of memory-mapped Register and direct memory access.
 */

#include <iostream>                 // std::cout
#include <sstream>                  // std::ostringstream
#include <iomanip>                  // std::setprecision

#include "Cstdint.h"                // std::uint32_t
#include "Register.h"               // class Register
#include "PerformanceTimer.h"       // class Register
#include "Test-common.h"            // SPM_TEST_INSPECT_MESSAGE_
#include <boost/test/unit_test.hpp> // Boost.Test

#define SILENCE_NO_ASSERT_MESSAGE  BOOST_CHECK( 1 )

/**
 * issue test code.
 */
#define TEST( code ) \
   setText( #code ); \
   timer.Start(); \
   for ( DataType x = k_loops ; x > 0; --x ) \
   { \
      code; \
      code; \
      code; \
      code; \
      code; \
      code; \
      code; \
      code; \
      code; \
      code; \
   } \
   timer.Stop(); \
   SILENCE_NO_ASSERT_MESSAGE;

typedef std::uint32_t DataType;
typedef spm::Register< DataType > RegisterType;

const int      k_N( 10 );           // # statements in loop
const DataType k_loops( 1000000 );  // # loops

std::string elapsed_ns( CPerformanceTimer /*const*/& timer )
{
   static int n = 0;

   std::ostringstream oss;
   oss << std::setw( 2 ) << ++n << "\t" <<
          std::setw( 4 ) << std::setprecision( 3 ) <<
          (1000.0 * timer.Elapsedus() / ( k_N * k_loops ));
   return oss.str();
}

BOOST_AUTO_TEST_SUITE( Register )
BOOST_AUTO_TEST_SUITE( Performance )

struct Fixture
{
   Fixture()
   : b()
   , cache()
   , data()
   , p(&data)
   , reg( &data, 0, 0 )
   {
      timer.Start();
   }

   ~Fixture()
   {
      BOOST_TEST_MESSAGE( Xelapsed_ns() << " [ns]\t" << text << "\t"<< SPM_TEST_CASE_NAME );
   }

   void setText( std::string text_ )
   {
      text = text_;
   }

   std::string Xelapsed_ns()
   {
      static int n = 0;

      std::ostringstream oss;
      oss << std::setw( 2 ) << ++n << "\t" <<
             std::setw( 4 ) << std::setprecision( 3 ) <<
             (1000.0 * timer.Elapsedus() / ( k_N * k_loops ));
      return oss.str();
   }

   volatile bool      b;
   volatile DataType  cache;
   volatile DataType  data;
   volatile DataType* p;
   RegisterType       reg;

   std::string        text;
   CPerformanceTimer  timer;

};

BOOST_FIXTURE_TEST_CASE( timePointerWriteNoCache, Fixture )
{
   TEST( *p = x )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterWriteNoCache, Fixture )
{
   TEST( reg.write_nc( x ) )
}


BOOST_FIXTURE_TEST_CASE( timePointerWriteAndCache, Fixture )
{
   TEST( *p = cache = x )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterWriteAndCache, Fixture )
{
   TEST( reg.write( x ) )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterAssignAndCache, Fixture )
{
   TEST( reg = x )
}


BOOST_FIXTURE_TEST_CASE( timePointerRead, Fixture )
{
   TEST( cache = *p )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterRead, Fixture )
{
   TEST( cache = reg.read() )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterReadOpData, Fixture )
{
   TEST( cache = reg )
}


BOOST_FIXTURE_TEST_CASE( timePointerBittest, Fixture )
{
   TEST( b = 0 != (*p & 0x01) )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterBittest, Fixture )
{
   TEST( b = reg.bittest( 0 ) )
}


BOOST_FIXTURE_TEST_CASE( timePointerBitsetFromCacheNoCache, Fixture )
{
   TEST( *p = (cache | 0x02) )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterBitsetFromCacheNoCache, Fixture )
{
   TEST( reg.write_nc( reg.cache() | 0x02 ) )
}


BOOST_FIXTURE_TEST_CASE( timePointerBitsetAndCache, Fixture )
{
   TEST( *p = cache |= 0x04 )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterBitsetAndCache, Fixture )
{
   TEST( reg.bitset( 3 ) )
}


BOOST_FIXTURE_TEST_CASE( timePointerMasksetFromCacheAndCache, Fixture )
{
   TEST( *p = cache |= 0x0f )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterMasksetAndCache, Fixture )
{
   TEST( reg.maskset( 0x0f ) )
}


BOOST_FIXTURE_TEST_CASE( timePointerMaskclearMasksetFromCacheAndCache, Fixture )
{
   TEST( *p = cache = ((cache & ~0xf0) | 0x30) )
}

BOOST_FIXTURE_TEST_CASE( timeRegisterMaskclearMasksetFromCacheAndCache, Fixture )
{
   TEST( reg.maskset( 0xf0, 0x30 ) )
}

BOOST_AUTO_TEST_SUITE_END() // Register
BOOST_AUTO_TEST_SUITE_END() // Performance

/*
 * end of file
 */
