/*
 * Code2/Register.h
 *
 * ACCU Overload 95, February 2010
 *
 * Supplement to: One approach to using hardware registers in C++.
 * Note 1: register read-write access control has been added.
 * Note 2: this version does not dependend on the Boost libraries.
 *
 * Martin Moene, http://www.eld.physics.LeidenUniv.nl/~moene/
 */

#ifndef G_REGISTER_H
#define G_REGISTER_H

/**
 * \file Code2/Register.h
 * A register class, transport via channel policy; transport for and creation
 * of memory mapped registers provided.
 */

// system includes
#include <memory>                         // std::auto_ptr
//#include <boost/noncopyable.hpp>          // boost::noncopyable
//#include <boost/static_assert.hpp>        // BOOST_STATIC_ASSERT()
//#include <boost/type_traits/is_same.hpp>  // boost::is_same
//#include <boost/type_traits/remove_cv.hpp>// boost::remove_cv

// project includes
#include "../Code/MemoryChannel.h"          // spm::MemoryChannel

/**
 * scanning probe microscopy.
 */
namespace spm {

/**
 * \ref Register2 "Register" access policies.
 */
namespace rap {

/// read-only register access.
struct ReadOnly {};

/// write-only register access.
struct WriteOnly {};

/// read-write register access.
struct ReadWrite : public ReadOnly, public WriteOnly {};

} // namespace rap

/**
 * \example Test-access.cpp
 * The following code shows how you can use the \ref rap "register access tags"
 * to get a register with read-only, write-only or read-write access.
 */

/**
 * A \ref Register2 "Register"; transport policy (for example memory mapped, or bus)
 * and register access policy are provided as template type parameters.
 *
 * Example:
 * \code
 * typedef int                        DataType;
 * typedef spm::Register2< DataType > RegisterType;
 * typedef RegisterType::AddressType  AddressType;
 *
 * int x;                        // simulate hardware register
 * const AddressType base  ( &x );
 * const int         offset(  0 );
 *
 * RegisterType reg( base, offset );
 *
 * reg = 456;                    // write given value to memory
 * reg.write();                  // write cached value to memory again
 * reg.write_nc( 987 );          // write given value to memory, but do not cache its value
 *
 * bool b = reg.bittest( 3 );    // test if bit 3 is set
 * reg.bitclear( 0 );            // clear first bit
 *
 * reg.maskclear( 0xf00 );       // clear bits according to given mask
 * reg.maskset( 0x0ff, 0x00f );  // clear bits in first mask, set bits in second mask
 * \endcode
 *
 * How exactly the register is read and written is left to a channel policy CP (or
 * transport policy). The access policy AP controls the kind of allowed access
 * to the register, such as read-only, write-only and read-write.
 * See \ref spm::rap "register access tags".
 * \code
 * template
 * < typename D
 * , typename A = volatile D*
 * , typename CP = spm::MemoryChannel<D,A>
 * , typename AP = spm::rap::ReadWrite
 * >
 * class Register2;
 * \endcode
 *
 * Here is how you define a read-only register.
 * \code
 * typedef spm::MemoryChannel< int >  RwChannelType;
 * typedef RwChannelType::DataType    DataType;
 * typedef RwChannelType::AddressType AddressType;
 *
 * typedef const spm::Register2< DataType, AddressType, RwChannelType, spm::rap::ReadOnly > RoRegisterType;
 *
 * DataType    x;                // simulate hardware register
 * AddressType base  ( &x );
 * int         offset(   0 )
 * DataType    value ( 123 );
 *
 * RoRegisterType roRegister( base, offset, value );
 * \endcode
 */
template
< typename D
, typename A = volatile D*
, typename CP = MemoryChannel<D,A>
, typename AP = rap::ReadWrite
>
class Register2 // : public boost::noncopyable
{
public:
   /**
    * the register data type.
    */
   typedef D DataType;

   /**
    * the register address type.
    */
   typedef A AddressType;

   /**
    * the transport channel type.
    */
   typedef CP ChannelType;

   /**
    * the access policy type.
    */
   typedef AP AccessType;

   /**
    * this type.
    */
   typedef Register2 Self;

   /**
    * constructor without channel object parameter.
    *
    * Note that although a register just has an address and the offset only makes
    * sense in relation the a card's base address, offset is included here to
    * confine the register's address computation to a single location.
    */
   Register2( AddressType address, int offset, DataType data = 0 )
   : m_channel_smartptr()           // Note: declared before m_channel
   , m_channel( createChannel() )
   , m_address( computeAddress( address, offset ) )
   , m_cache( data )
   {
      ; // do nothing
   }

   /**
    * constructor with channel object parameter.
    *
    * Note that although a register just has an address and the offset only makes
    * sense in relation the a card's base address, offset is included here to
    * confine the register's address computation to a single location.
    */
   Register2( ChannelType& channel, AddressType address, int offset, DataType data = 0 )
   : m_channel_smartptr()           // Note: declared before m_channel
   , m_channel( channel )
   , m_address( computeAddress( address, offset ) )
   , m_cache  ( data    )
   {
      ; // do nothing
   }

   /**
    * destructor: used for type checking.
    */
   ~Register2()
   {
      checkTypes();
   }

   /**
    * ensure that data and address types of channel policy and register match.
    */
   static void checkTypes()
   {
//      BOOST_STATIC_ASSERT( ( boost::is_same<D, typename CP::DataType>::value ) );
//      BOOST_STATIC_ASSERT( ( boost::is_same<A, typename CP::AddressType>::value ) );
   }

   /**
    * read data and return it; same as read().
    */
   operator DataType() const
   {
      return read();
   }

   /**
    * write data and return self; in effect write(data).
    */
   Self& operator= ( DataType data )
   {
      write( data );
      return *this;
   }

   /**
    * reinitialize the register on the card from the cached value.
    */
   void reinitialize()
   {
      write();
   }

   /**
    * the register's address.
    */
   AddressType address() const
   {
      return m_address;
   }

   /**
    * the register's cached data (non-reading).
    */
   DataType cache() const
   {
      return m_cache;
   }

   /**
    * read the register and return its value (reading).
    */
   DataType read() const
   {
      return read( AccessType() );
   }

   /**
    * write the cached value to the register on the card.
    */
   void write()
   {
      write( AccessType() );
   }

   /**
    * write the given value to the register (caching).
    */
   void write( DataType value )
   {
      write( value, AccessType() );
   }

   /**
    * write the given value to the register (non-caching).
    */
   void write_nc( DataType value )
   {
      write_nc( value, AccessType() );
   }

   /**
    * test the bit at the given position of the register (reading).
    */
   bool bittest( int bit ) const
   {
      return 0 != ( read() & bitmask( bit ) );
   }

   /**
    * test the bit at the given position of the register (cached).
    */
   bool bittest_c( int bit ) const
   {
      return 0 != ( cache() & bitmask( bit ) );
   }

   /**
    * clear the bit at the given position of the register (caching).
    */
   void bitclear( int bit )
   {
      write( m_cache & ~bitmask( bit ) );
   }

   /**
    * set the bit at the given position of the register (caching).
    */
   void bitset( int bit )
   {
      write( m_cache | bitmask( bit ) );
   }

   /**
    * test the bits in the register at the positions given by the mask (reading).
    */
   bool masktest( DataType mask ) const
   {
      return mask == ( read() & mask );
   }

   /**
    * test the bits in the register at the positions given by the mask (cached).
    */
   bool masktest_c( DataType mask ) const
   {
      return mask == ( cache() & mask );
   }

   /**
    * clear the bits in the register at the positions given by the mask (caching).
    */
   void maskclear( DataType mask )
   {
      write( m_cache & ~mask );
   }

   /**
    * set the bits in the register at the positions given by the mask (caching).
    */
   void maskset( DataType mask )
   {
      write( m_cache | mask );
   }

   /**
    * in the register, clear the bits given by clearmask, set the bits given by the setmask (caching).
    */
   void maskset( DataType clearmask, DataType setmask )
   {
      m_cache &= ~clearmask;
      maskset( setmask );
   }

   /**
    * create bit mask for given bit.
    */
   static DataType bitmask( int bit )
   {
      return 1 << bit;
   }

   /**
    * compute register address method.
    */
   static AddressType computeAddress( AddressType base, int offset )
   {
      return base + offset;
   }

private:
   /**
    * prevent copy-initialisation.
    */
   Register2( Self const& rhs );

   /**
    * prevent copy-assignment.
    */
   Self& operator=( Self const& rhs );

   /**
    * read the register and return its value (reading, ReadOnly access).
    */
   DataType read( rap::ReadOnly ) const
   {
      return m_channel.read( m_address );
   }

   /**
    * write the cached value to the register on the card (WriteOnly access).
    */
   void write( rap::WriteOnly )
   {
      m_channel.write( m_address, m_cache );
   }

   /**
    * write the given value to the register (caching, WriteOnly access).
    */
   void write( DataType value, rap::WriteOnly )
   {
      m_channel.write( m_address, m_cache = value );
   }

   /**
    * write the given value to the register (non-caching, WriteOnly access).
    */
   void write_nc( DataType value, rap::WriteOnly )
   {
      m_channel.write( m_address, value );
   }

   /**
    * constructor helper method.
    */
   ChannelType& createChannel()
   {
      m_channel_smartptr.reset( new ChannelType() );
      return *m_channel_smartptr;
   }

private:
   /*
    * Note1: std::auto_ptr must be declared before m_channel (init before use).
    * Note2: std::auto_ptr is to be replaced by std::unique_ptr (C++0X).
    */

   /**
    * smart pointer for internally created channels, such as transport for memory mapped registers.
    */
   std::auto_ptr<ChannelType> m_channel_smartptr;

   /**
    * the transport channel policy.
    */
   ChannelType& m_channel;

   /**
    * the register's address.
    */
   AddressType m_address;

   /**
    * the cached written data.
    */
   DataType m_cache;
};

} // namespace spm {

#endif // G_REGISTER_H

/*
 * end of file
 */
